r"""Blocks of network layers with residual skip connections."""

from __future__ import annotations

from typing import Any, Mapping, Optional, Union

import torch
from torch.nn import Identity, Module, Sequential, init

from deepali.core.enum import PaddingMode
from deepali.core.typing import ScalarOrTuple

from ..layers.acti import ActivationArg, activation
from ..layers.conv import ConvLayer, convolution, same_padding
from ..layers.norm import NormArg
from ..layers.join import JoinLayer

from .skip import SkipConnection, SkipFunc


# fmt: off
__all__ = (
    "ResidualUnit",
)
# fmt: on


class ResidualUnit(SkipConnection):
    r"""Sequence of convolutional layers with a residual skip connection.

    Implements a number of variants of residual blocks as described in:
    - He et al., 2015, Deep Residual Learning for Image Recognition, https://arxiv.org/abs/1512.03385
    - He et al., 2016, Identity Mappings in Deep Residual Networks, https://arxiv.org/abs/1603.05027
    - Xie et al., 2017, Aggregated Residual Transformations for Deep Neural Networks, https://arxiv.org/abs/1611.05431

    """

    def __init__(
        self,
        spatial_dims: int,
        in_channels: int,
        out_channels: Optional[int] = None,
        num_channels: Optional[int] = None,
        kernel_size: ScalarOrTuple[int] = 3,
        stride: ScalarOrTuple[int] = 1,
        padding: Optional[ScalarOrTuple[int]] = None,
        padding_mode: Union[PaddingMode, str] = "zeros",
        dilation: ScalarOrTuple[int] = 1,
        groups: int = 1,
        init: str = "default",
        bias: Optional[Union[bool, str]] = False,
        norm: NormArg = "batch",
        acti: ActivationArg = "relu",
        skip: Optional[Union[SkipFunc, str, Mapping[str, Any]]] = "identity | conv1 | conv",
        num_layers: Optional[int] = None,
        order: str = "cna",
        pre_conv: str = "conv1",
        post_conv: str = "conv1",
        other: Optional[ResidualUnit] = None,
    ) -> None:
        r"""Initialize residual unit.

        Args:
            spatial_dims: Number of spatial input and output dimensions.
            in_channels: Number of input feature maps.
            out_channels: Number of output feature maps.
            num_channels: Number of input feature maps to spatial convolutions.
                If ``None`` or equal to ``out_channels``, the residual branch consists of ``num_layers``
                of convolutions with specified ``kernel_size`` and pre- or post-activation (and normalization).
                Otherwise, the first and last convolution has kernel size 1 in order to match the
                specified number of input and output channels of the shortcut connection.
                When the interim number of channels is smaller than the input and output channels,
                this residual unit is a so-called bottleneck block.
            kernel_size: Size of residual convolutions.
            pre_conv: Type of pre-convolution when residual unit is a bottleneck block.
                The kernel size is set to 1 if "conv1", and ``kernel_size`` otherwise.
            post_conv: Type of post-convolution when residual unit is a bottleneck block. See ``pre_conv``.
            stride: Stride of first residual convolution. All subsequent convolutions have stride 1.
                In case of a bottleneck block, the first convolution with kernel size 1 also has stride 1,
                and the specified stride is applied on the second convolution instead.
            padding: Padding used by residual convolutions with specified ``kernel_size``.
                If specified, must result in a same padding such that spatial tensor shape remains unchanged.
            padding_mode: Padding mode used for same padding of input to each convolutional layers.
            dilation: Dilation of convolution kernel.
            groups: Number of groups into which to split each convolution. Note that combined with a bottleneck,
                i.e., where ``num_channels`` is smaller than ``out_channels``, this is equivalent to increasing
                the number of residual paths also referred to as cardinality (cf. ResNeXt). For example, setting
                ``in_channels=256``, out_channels=256``, ``num_channels=128``, and ``groups=32`` is equivalent to
                a residual block consisting of 32 paths with 4 channels in each path which are concatentated prior
                to the final convolution with kernel size 1 before the resulting residuals are added to the result
                of the (identity) skip connection (cf. Xie et al., 2017, Fig. 3).
            init: Mode used to initialize weights of convolution kernels.
            bias: Whether to use bias terms of convolutional filters.
            norm: Normalization layer to use in each convolutional layer.
            acti: Activation function to use in each convolutional layer.
            skip: Function along the shortcut connection. If specified, must ensure that the
                shape of the output along this shortcut connection is identical to the shape of
                the output from the sequence of convolutional layers in this block. If ``None``,
                the shortcut connection is the identity map. By default, the identity map is used when the
                output shape matches the input shape. Otherwise, a convolution with kernel size 1 is used if
                ``in_channels != out_channels``, or a convolution with specified ``kernel_size``. The argument
                can be a string specifying which of these, i.e., "identity", "conv1", or "conv", to consider.
                Multiple options can be allowed by using the "|" operator. The default is "identity | conv1 | conv".
                In order to always use a convolution with kernel size 1, even in case of a stride > 1, use
                "identity | conv1". To force the use of the identity, set argument to ``None`` or "identity".
            num_layers: Number of convolutional layers in residual block.
            order: Order of convolution, normalization, and activation in each layer.
                In case of post-convolution activation (post-activiation), the final activation
                is applied after the addition of the layer output with the shortcut connection.
                Otherwise, the addition is of input and output is the last operation of this block.
            other: If specified, convolutions in convolutional layers are reused. The given residual unit
                must have created with the same parameters as this residual unit. Note that the resulting
                and the other residual unit will share references to the same convolution modules.

        """
        if spatial_dims < 0 or spatial_dims > 3:
            raise ValueError("ResidualUnit() 'spatial_dims' must be 1, 2, or 3")
        order = order.upper()
        if len(set(order)) != len(order) or "C" not in order or "A" not in order:
            raise ValueError(
                "ResidualUnit() 'order' must be permutation of unique characters"
                f" 'a|A' (activation), 'n|N' (norm, optional), and 'c|C' (convolution), got {order!r}"
            )
        # Pre- vs. post-activation
        if order.index("C") < order.index("A"):
            post_acti = activation(acti)
        else:
            post_acti = None
        # Add convolutional layers
        if out_channels is None:
            out_channels = in_channels
        if num_channels is None:
            num_channels = out_channels
        is_bottleneck = num_channels != out_channels
        if num_layers is None:
            num_layers = 3 if is_bottleneck else 2
        if not isinstance(num_layers, int):
            raise TypeError("ResidualUnit() 'num_layers' must be int or None")
        if is_bottleneck and num_layers < 3:
            raise ValueError(
                "ResidualUnit() 'num_layers' must be at least 3 in case of a bottleneck block"
            )
        elif num_layers < 1:
            raise ValueError("ResidualUnit() 'num_layers' must be positive")
        residual = Sequential()
        if is_bottleneck:
            name = f"layer_{len(residual)}"
            if other is None:
                other_conv = None
            else:
                other_layer = getattr(other.residual, name)
                assert isinstance(other_layer, ConvLayer)
                other_conv = other_layer.conv
                assert isinstance(other_conv, Module)
            if pre_conv == "conv":
                pre_kernel_size = kernel_size
            elif pre_conv == "conv1":
                pre_kernel_size = 1
            else:
                raise ValueError("ResidualUnit() 'pre_conv' must be 'conv' or 'conv1'")
            conv = ConvLayer(
                spatial_dims=spatial_dims,
                in_channels=in_channels,
                out_channels=num_channels,
                kernel_size=pre_kernel_size,
                padding_mode=padding_mode,
                init=init,
                bias=bias,
                norm=norm,
                acti=acti,
                order=order,
                conv=other_conv,
            )
            residual.add_module(name, conv)
        for i in range(1 if is_bottleneck else 0, num_layers - 1 if is_bottleneck else num_layers):
            name = f"layer_{len(residual)}"
            is_first_layer = i == 0
            is_last_layer = i == num_layers - 1
            if other is None:
                other_conv = None
            else:
                other_layer = getattr(other.residual, name)
                assert isinstance(other_layer, ConvLayer)
                other_conv = other_layer.conv
                assert isinstance(other_conv, Module)
            conv = ConvLayer(
                spatial_dims=spatial_dims,
                in_channels=in_channels if is_first_layer else num_channels,
                out_channels=num_channels,
                kernel_size=kernel_size,
                stride=stride if is_first_layer else 1,
                padding=padding,
                padding_mode=padding_mode,
                dilation=dilation,
                groups=groups,
                init=init,
                bias=bias,
                norm=norm,
                acti=None if is_last_layer and post_acti is not None else acti,
                order=order,
                conv=other_conv,
            )
            residual.add_module(name, conv)
        if is_bottleneck:
            name = f"layer_{len(residual)}"
            if other is None:
                other_conv = None
            else:
                other_layer = getattr(other.residual, name)
                assert isinstance(other_layer, ConvLayer)
                other_conv = other_layer.conv
                assert isinstance(other_conv, Module)
            if post_conv == "conv":
                post_kernel_size = kernel_size
            elif post_conv == "conv1":
                post_kernel_size = 1
            else:
                raise ValueError("ResidualUnit() 'post_conv' must be 'conv' or 'conv1'")
            conv = ConvLayer(
                spatial_dims=spatial_dims,
                in_channels=num_channels,
                out_channels=out_channels,
                kernel_size=post_kernel_size,
                padding_mode=padding_mode,
                init=init,
                bias=bias,
                norm=norm,
                acti=acti if post_acti is None else None,
                order=order,
                conv=other_conv,
            )
            residual.add_module(name, conv)
        # Apply convolution along skip connection to match shape of residual tensor
        has_strided_conv = not torch.as_tensor(stride, dtype=torch.int, device="cpu").prod().eq(1)
        if skip in (None, ""):
            skip = "identity"
        if isinstance(skip, str):
            skip_names = [s.strip() for s in skip.lower().split("|")]
            if has_strided_conv or in_channels != out_channels:
                if "conv" in skip_names and (has_strided_conv or "conv1" not in skip_names):
                    skip = {}
                elif "conv1" in skip_names:
                    skip = {"kernel_size": 1, "padding": 0}
                elif skip == "identity":
                    raise ValueError(
                        "ResidualUnit() cannot use 'identity' skip connection"
                        f" (has_strided_conv={has_strided_conv}"
                        f", in_channels={in_channels}"
                        f", out_channels={out_channels})"
                    )
                else:
                    raise ValueError(f"ResidualUnit() invalid 'skip' value {skip!r}")
            elif "identity" in skip_names:
                skip = Identity()
            elif "conv1" in skip_names:
                skip = {"kernel_size": 1, "padding": 0}
            elif "conv" in skip_names:
                skip = {}
            else:
                raise ValueError(f"ResidualUnit() invalid 'skip' value {skip!r}")
        if isinstance(skip, dict):
            if "padding" in skip and "kernel_size" not in skip:
                raise ValueError("ResidualUnit() 'skip' specifies 'padding' but not 'kernel_size'")
            conv_args = dict(
                kernel_size=kernel_size if has_strided_conv else 1,
                stride=stride,
                dilation=1,
                init=init,
                bias=bias is not False,
            )
            conv_args.update(skip)
            if "padding" not in conv_args:
                conv_args["padding"] = same_padding(conv_args["kernel_size"], conv_args["dilation"])
            skip = convolution(spatial_dims, in_channels, out_channels, **conv_args)
        elif not callable(skip):
            raise TypeError("ResidualUnit() 'skip' must be str, dict, callable, or None")
        # Function used to combine input with residual tensor
        add = JoinLayer("add")
        if post_acti is None:
            join = add
        else:
            join = Sequential()
            join.add_module("add", add)
            join.add_module("acti", post_acti)
        # Initialize parent class and set module attributes
        super().__init__(residual, name="residual", skip=skip, join=join)
        self.spatial_dims = spatial_dims
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.num_channels = num_channels

    def is_bottleneck(self) -> bool:
        r"""Whether this residual unit is a "bottleneck" type ResNet block."""
        return self.out_channels != self.num_channels

    @property
    def last_conv_layer(self) -> ConvLayer:
        r"""Get last residual convolutional layer."""
        return self.func[-1]

    def zero_init_residual(self) -> None:
        r"""Zero-initialize normalization layer in residual branch.

        This is so that the residual branch starts with zeros, and each residual block behaves like an identity.
        This improves the model by 0.2~0.3% according to https://arxiv.org/abs/1706.02677.

        """
        conv_layer = self.last_conv_layer
        if conv_layer.has_norm_after_conv():
            weight = getattr(conv_layer.norm, "weight", None)
            init.constant_(weight, 0)



r"""Basic network modules, usually with learnable parameters."""

from deepali.modules import Pad
from deepali.modules import Reshape
from deepali.modules import View

from .acti import Activation
from .acti import ActivationArg
from .acti import ActivationFunc
from .acti import activation
from .acti import is_activation

from .conv import Conv1d
from .conv import Conv2d
from .conv import Conv3d
from .conv import ConvTranspose1d
from .conv import ConvTranspose2d
from .conv import ConvTranspose3d
from .conv import ConvLayer
from .conv import conv_module
from .conv import convolution
from .conv import is_convolution
from .conv import is_conv_module

from .join import JoinLayer
from .join import JoinFunc
from .join import join_func

from .lambd import LambdaLayer
from .lambd import LambdaFunc

from .linear import Linear

from .norm import NormArg
from .norm import NormFunc
from .norm import NormLayer
from .norm import norm_layer
from .norm import normalization
from .norm import is_batch_norm
from .norm import is_group_norm
from .norm import is_instance_norm
from .norm import is_norm_layer

from .pool import PoolArg
from .pool import PoolFunc
from .pool import PoolLayer
from .pool import pool_layer
from .pool import pooling

from .upsample import Upsample
from .upsample import UpsampleMode
from .upsample import SubpixelUpsample


__all__ = (
    # Basic layers
    "Pad",
    "Reshape",
    "View",
    # Non-linear activations
    "Activation",
    "ActivationArg",
    "ActivationFunc",
    "activation",
    "is_activation",
    # Convolutional layers
    "Conv1d",
    "Conv2d",
    "Conv3d",
    "ConvTranspose1d",
    "ConvTranspose2d",
    "ConvTranspose3d",
    "ConvLayer",
    "conv_module",
    "convolution",
    "is_convolution",
    "is_conv_module",
    # Merge network branches
    "JoinLayer",
    "JoinFunc",
    "join_func",
    # Fully connected layer
    "Linear",
    # Normalization layers
    "NormArg",
    "NormFunc",
    "NormLayer",
    "norm_layer",
    "normalization",
    "is_batch_norm",
    "is_group_norm",
    "is_instance_norm",
    "is_norm_layer",
    # Pooling layers
    "PoolArg",
    "PoolFunc",
    "PoolLayer",
    "pool_layer",
    "pooling",
    # Resampling layers
    "Upsample",
    "UpsampleMode",
    "SubpixelUpsample",
    # Misc layers
    "LambdaLayer",
    "LambdaFunc",
)



r"""Functions for preparing model input containing tensor decorators."""

from collections import abc
from dataclasses import is_dataclass
from typing import Any, Mapping, NamedTuple, Optional, Sequence, Union, overload

import torch
from torch import Tensor

from deepali.core.typing import Batch, Dataclass, Device, is_namedtuple

from .sample import sample_field_names, sample_field_value
from .sample import replace_all_sample_field_values


__all__ = ("prepare_batch",)


@overload
def prepare_batch(batch: Sequence[Mapping[str, Any]]) -> Mapping[str, Any]:
    ...


@overload
def prepare_batch(batch: Sequence[Dataclass]) -> Dataclass:
    ...


@overload
def prepare_batch(batch: Sequence[NamedTuple]) -> NamedTuple:
    ...


def prepare_batch(
    batch: Batch,
    device: Optional[Union[Device, str]] = None,
    non_blocking: bool = False,
    memory_format=torch.preserve_format,
) -> Batch:
    r"""Move batch data to execution device."""
    names = sample_field_names(batch)
    values = []
    for name in names:
        value = sample_field_value(batch, name)
        value = prepare_item(
            value, device=device, non_blocking=non_blocking, memory_format=memory_format
        )
        values.append(value)
    return replace_all_sample_field_values(batch, values)


def prepare_item(
    value: Any,
    device: Optional[Union[Device, str]] = None,
    non_blocking: bool = False,
    memory_format=torch.preserve_format,
) -> Any:
    r"""Move batch item data to execution device."""
    kwargs = dict(device=device, non_blocking=non_blocking, memory_format=memory_format)
    if isinstance(value, Tensor):
        value = value.to(**kwargs)
    elif isinstance(value, abc.Mapping) or is_dataclass(value) or is_namedtuple(value):
        value = prepare_batch(value, **kwargs)
    elif isinstance(value, Sequence) and not isinstance(value, str):
        value = [prepare_item(item, **kwargs) for item in value]
    return value



from deepali.core.pathlib import PathUri, unlink_or_mkdir
from deepali.core.storage import StorageObject

import SimpleITK as sitk


def read_image(path: PathUri) -> sitk.Image:
    r"""Read image from file."""
    with StorageObject.from_path(path) as obj:
        if not obj.is_file():
            raise FileNotFoundError(f"No such file or no access: '{path}'")
        obj.pull(force=True)
        image = sitk.ReadImage(str(obj.path))
    return image


def write_image(image: sitk.Image, path: PathUri, compress: bool = True):
    r"""Write image to file."""
    with StorageObject.from_path(path) as obj:
        path = unlink_or_mkdir(obj.path)
        sitk.WriteImage(image, str(path), compress)
        obj.push(force=True)



r"""Output conversion modules."""

from typing import Mapping, Sequence, Union

from torch import Tensor
from torch.nn import Module

from deepali.core.nnutils import as_immutable_container


class ToImmutableOutput(Module):
    r"""Convert input to immutable output container.

    For use with ``torch.utils.tensorboard.SummaryWriter.add_graph`` when model output is list or dict.
    See error message: "Encountering a dict at the output of the tracer might cause the trace to be incorrect,
    this is only valid if the container structure does not change based on the module's inputs. Consider using
    a constant container instead (e.g. for `list`, use a `tuple` instead. for `dict`, use a `NamedTuple` instead).
    If you absolutely need this and know the side effects, pass strict=False to trace() to allow this behavior."

    """

    def __init__(self, recursive: bool = True) -> None:
        super().__init__()
        self.recursive = recursive

    def forward(self, input: Union[Tensor, Sequence, Mapping]) -> Union[Tensor, Sequence, Mapping]:
        return as_immutable_container(input, recursive=self.recursive)

    def extra_repr(self) -> str:
        return f"recursive={self.recursive!r}"



import torch
from torch import Tensor, nn
from torch.nn import Module

from deepali.core.nnutils import conv_output_size, conv_transposed_output_size
from deepali.core.nnutils import pad_output_size, upsample_output_size
from deepali.core.nnutils import pool_output_size, unpool_output_size
from deepali.core.typing import ScalarOrTuple

from .layers import Pad
from .blocks import SkipConnection
from .layers import is_activation
from .layers import is_norm_layer


def module_output_size(module: Module, in_size: ScalarOrTuple[int]) -> ScalarOrTuple[int]:
    r"""Calculate spatial size of output tensor after the given module is applied."""
    if not isinstance(module, Module) or type(module) is Module:
        raise TypeError("module_output_size() 'module' must be torch.nn.Module subclass")
    # Modules defining an output_size attribute or function
    output_size = getattr(module, "output_size", None)
    if callable(output_size):
        return output_size(in_size)
    if output_size is not None:
        device = torch.device("cpu")
        m: Tensor = torch.atleast_1d(torch.tensor(in_size, dtype=torch.int, device=device))
        if m.ndim != 1:
            raise ValueError("module_output_size() 'in_size' must be scalar or sequence")
        ndim = m.shape[0]
        s: Tensor = torch.atleast_1d(torch.tensor(output_size, dtype=torch.int, device=device))
        if s.ndim != 1 or s.shape[0] not in (1, ndim):
            raise ValueError(
                f"module_output_size() 'module.output_size' must be scalar or sequence of length {ndim}"
            )
        n = s.expand(ndim)
        if isinstance(in_size, int):
            return n.item()
        return tuple(n.tolist())
    # Network blocks
    if isinstance(module, nn.Sequential):
        size = in_size
        for m in module:
            size = module_output_size(m, size)
        return size
    if isinstance(module, SkipConnection):
        return module_output_size(module.func, in_size)
    # Convolutional layers
    if isinstance(module, (nn.Conv1d, nn.Conv2d, nn.Conv3d)):
        return conv_output_size(
            in_size,
            kernel_size=module.kernel_size,
            stride=module.stride,
            padding=module.padding,
            dilation=module.dilation,
        )
    if isinstance(module, (nn.ConvTranspose1d, nn.ConvTranspose2d, nn.ConvTranspose3d)):
        return conv_transposed_output_size(
            in_size,
            kernel_size=module.kernel_size,
            stride=module.stride,
            padding=module.padding,
            output_padding=module.output_padding,
            dilation=module.dilation,
        )
    # Pooling layers
    if isinstance(
        module,
        (
            nn.AvgPool1d,
            nn.AvgPool2d,
            nn.AvgPool3d,
            nn.MaxPool1d,
            nn.MaxPool2d,
            nn.MaxPool3d,
        ),
    ):
        return pool_output_size(
            in_size,
            kernel_size=module.kernel_size,
            stride=module.stride,
            padding=module.padding,
            dilation=module.dilation,
            ceil_mode=module.ceil_mode,
        )
    if isinstance(
        module,
        (
            nn.AdaptiveAvgPool1d,
            nn.AdaptiveAvgPool2d,
            nn.AdaptiveAvgPool3d,
            nn.AdaptiveMaxPool1d,
            nn.AdaptiveMaxPool2d,
            nn.AdaptiveMaxPool3d,
        ),
    ):
        return module.output_size
    if isinstance(module, (nn.MaxUnpool1d, nn.MaxUnpool2d, nn.MaxUnpool3d)):
        return unpool_output_size(
            in_size,
            kernel_size=module.kernel_size,
            stride=module.stride,
            padding=module.padding,
        )
    # Padding
    if isinstance(module, Pad):
        raise NotImplementedError()
    if isinstance(
        module,
        (
            nn.ReflectionPad1d,
            nn.ReflectionPad2d,
            nn.ReplicationPad1d,
            nn.ReplicationPad2d,
            nn.ReplicationPad3d,
            nn.ZeroPad2d,
            nn.ConstantPad1d,
            nn.ConstantPad2d,
            nn.ConstantPad3d,
        ),
    ):
        return pad_output_size(in_size, module.padding)
    # Upsampling
    if isinstance(module, nn.Upsample):
        return upsample_output_size(in_size, size=module.size, scale_factor=module.scale_factor)
    # Activation functions
    if is_activation(module) or isinstance(
        module,
        (
            nn.ELU,
            nn.Hardshrink,
            nn.Hardsigmoid,
            nn.Hardtanh,
            nn.Hardswish,
            nn.LeakyReLU,
            nn.LogSigmoid,
            nn.LogSoftmax,
            nn.PReLU,
            nn.ReLU,
            nn.ReLU6,
            nn.RReLU,
            nn.SELU,
            nn.CELU,
            nn.GELU,
            nn.Sigmoid,
            nn.Softmax,
            nn.Softmax2d,
            nn.Softmin,
            nn.Softplus,
            nn.Softshrink,
            nn.Softsign,
            nn.Tanh,
            nn.Tanhshrink,
            nn.Threshold,
        ),
    ):
        return in_size
    # Normalization layers
    if is_norm_layer(module) or isinstance(
        module,
        (
            nn.BatchNorm1d,
            nn.BatchNorm2d,
            nn.BatchNorm3d,
            nn.SyncBatchNorm,
            nn.GroupNorm,
            nn.InstanceNorm1d,
            nn.InstanceNorm2d,
            nn.InstanceNorm3d,
            nn.LayerNorm,
            nn.LocalResponseNorm,
        ),
    ):
        return in_size
    # Dropout layers
    if isinstance(module, (nn.AlphaDropout, nn.Dropout, nn.Dropout2d, nn.Dropout3d)):
        return in_size
    # Not implemented or invalid type
    if isinstance(module, (nn.ModuleDict, nn.ModuleList)):
        raise TypeError(
            "module_output_size() order of modules in ModuleDict or ModuleList is undetermined"
        )
    if isinstance(module, (nn.ParameterDict, nn.ParameterList)):
        raise TypeError(
            "module_output_size() 'module' cannot be torch.nn.ParameterDict or torch.nn.ParameterList"
        )
    raise NotImplementedError(
        f"module_output_size() not implemented for 'module' of type {type(module)}"
    )



import torch

from deepali.core import functional as U
from deepali.data import ImageBatch
from deepali.spatial import Translation
from deepali.spatial import ImageTransformer


def test_spatial_image_transformer() -> None:
    # Generate sample image
    image = ImageBatch(U.cshape_image(size=(65, 33), center=(32, 16), sigma=1, dtype=torch.float32))
    assert image.shape == (1, 1, 33, 65)

    # Apply identity transformation
    offset = torch.tensor([0, 0], dtype=torch.float32)
    translation = Translation(image.grid(), params=offset.unsqueeze(0))
    transformer = ImageTransformer(translation)
    warped = transformer.forward(image)
    assert warped.eq(image).all()

    # Shift image in world (and image) space to the left
    offset = torch.tensor([0.5, 0], dtype=torch.float32)
    translation = Translation(image.grid(), params=offset.unsqueeze(0))
    transformer = ImageTransformer(translation)
    warped = transformer.forward(image)
    expected = U.cshape_image(size=(65, 33), center=(32 - 16, 16), sigma=1, dtype=image.dtype)
    assert torch.allclose(warped, expected)

    # Shift image in world (and image) space to the right
    offset = torch.tensor([-0.5, 0], dtype=torch.float32)
    translation = Translation(image.grid(), params=offset.unsqueeze(0))
    transformer = ImageTransformer(translation)
    warped = transformer.forward(image)
    expected = U.cshape_image(size=(65, 33), center=(32 + 16, 16), sigma=1, dtype=image.dtype)
    assert torch.allclose(warped, expected)

    # Shift target sampling grid in world space and sample input with identity transform.
    # This results in a shift of the image in the image space though world positions are unchanged.
    target = image.grid().center(32 - 16, 0)
    assert not target.same_domain_as(image.grid())

    assert torch.allclose(
        image.grid().index_to_world([32, 16]),
        target.index_to_world([16, 16]),
    )

    offset = torch.tensor([0, 0], dtype=torch.float32)
    translation = Translation(image.grid(), params=offset.unsqueeze(0))
    transformer = ImageTransformer(translation, target=target, source=image.grid())
    warped = transformer.forward(image)
    expected = U.cshape_image(size=(65, 33), center=(32 - 16, 16), sigma=1, dtype=image.dtype)
    assert torch.allclose(warped, expected)



r"""Mix-ins for spatial transformations that have (optimizable) parameters.

These `mix-ins <https://en.wikipedia.org/wiki/Mixin>`_ add property ``params`` to a
:class:`.SpatialTransform`, which can be either one of the following. In addition, functional
setter and getter functions are added, which check the type and shape of its arguments.

- ``None``: This spatial transformation has no required parameters set.
    This value can be specified when initializing a spatial transformation whose parameters
    will be set at a later time point, e.g., to the output of a neural network. An exception
    is raised by functions which attempt to access yet uninitialized transformation parameters.
- ``Parameter``: Tensor of optimizable parameters, e.g., for classic registration.
    To temporarily disable optimization of the parameters, set ``params.requires_grad = False``.
- ``Tensor``: Tensor of fixed non-optimizable parameters.
    These parameters are not returned by :meth:`.SpatialTransform.parameters`. This could be a
    tensor of spatial transformation parameters inferred by a neural network.
- ``Callable``: A callable such as a function or ``torch.nn.Module``.
    Method :meth:`.SpatialTransform.update`, which is registered as pre-forward hook for any spatial
    transformation, invokes this callable to obtain the current transformation parameters with arguments
    set and obtained by :meth:`.SpatialTransform.condition`. For example, an input batch of a neural
    network can be passed to a ``torch.nn.Module`` this way to infer parameters from this input.

"""

from __future__ import annotations

from copy import copy as shallow_copy
from typing import Callable, Optional, Union, cast, overload

import torch
from torch import Size, Tensor
from torch.nn import Parameter, init

from deepali.core.grid import Grid

from .base import ReadOnlyParameters, TSpatialTransform


class ParametricTransform:
    r"""Mix-in for spatial transformations that have (optimizable) parameters."""

    def __init__(
        self: Union[TSpatialTransform, ParametricTransform],
        grid: Grid,
        groups: Optional[int] = None,
        params: Optional[Union[bool, Tensor, Callable]] = True,
    ) -> None:
        r"""Initialize transformation parameters.

        Args:
            grid: Grid domain on which transformation is defined.
            groups: Number of transformations. A given image batch can either be deformed by a
                single transformation, or a separate transformation for each image in the batch, e.g.,
                for group-wise or batched registration. The default is one transformation for all images
                in the batch, or the batch length of the ``params`` tensor if provided.
            params: Initial parameters. If a tensor is given, it is only registered as optimizable module
                parameters when of type ``torch.nn.Parameter``. When a callable is given instead, it will be
                called by ``self.update()`` with ``SpatialTransform.condition()`` arguments. When a boolean
                argument is given, a new zero-initialized tensor is created. If ``True``, this tensor is
                registered as optimizable module parameter. If ``None``, parameters must be set using
                ``self.data()`` or ``self.data_()`` before this transformation is evaluated.

        """
        if isinstance(params, Tensor) and params.ndim < 2:
            raise ValueError(
                f"{type(self).__name__}() 'params' tensor must be at least 2-dimensional"
            )
        super().__init__(grid)
        if groups is None:
            groups = params.shape[0] if isinstance(params, Tensor) else 1
        shape = (groups,) + self.data_shape
        if params is None:
            self.params = None
        elif isinstance(params, bool):
            data = torch.empty(shape, dtype=torch.float)
            if params:
                self.params = Parameter(data)
            else:
                self.register_buffer("params", data, persistent=True)
            self.reset_parameters()
        elif isinstance(params, Tensor):
            if shape and params.shape != shape:
                raise ValueError(
                    f"{type(self).__name__}() 'params' must be tensor of shape {shape!r}"
                )
            if isinstance(params, Parameter):
                self.params = params
            else:
                self.register_buffer("params", params, persistent=True)
        elif callable(params):
            self.params = params
            self.register_buffer("p", torch.empty(shape), persistent=False)
            self.reset_parameters()
        else:
            raise TypeError(
                f"{type(self).__name__}() 'params' must be bool, Callable, Tensor, or None"
            )

    def has_parameters(self) -> bool:
        r"""Whether this transformation has optimizable parameters."""
        return isinstance(self.params, Parameter)

    @torch.no_grad()
    def reset_parameters(self: Union[TSpatialTransform, ParametricTransform]) -> None:
        r"""Reset transformation parameters."""
        params = self.params  # Note: May be None!
        if params is None:
            return
        if callable(params):
            params = self.p
        init.constant_(params, 0.0)
        self.clear_buffers()

    @property
    def data_shape(self) -> Size:
        r"""Get required shape of transformation parameters tensor, excluding batch dimension."""
        raise NotImplementedError(f"{type(self).__name__}.data_shape")

    @overload
    def data(self) -> Tensor:
        r"""Get (buffered) transformation parameters."""
        ...

    @overload
    def data(self: TSpatialTransform, arg: Tensor) -> TSpatialTransform:
        r"""Get shallow copy with specified parameters."""
        ...

    def data(
        self: Union[TSpatialTransform, ParametricTransform], arg: Optional[Tensor] = None
    ) -> Union[TSpatialTransform, Tensor]:
        r"""Get transformation parameters or shallow copy with specified parameters, respectively."""
        params = self.params  # Note: May be None!
        if arg is None:
            if params is None:
                raise AssertionError(f"{type(self).__name__}.data() 'params' must be set first")
            if callable(params):
                params = getattr(self, "p")
            return params
        if not isinstance(arg, Tensor):
            raise TypeError(f"{type(self).__name__}.data() 'arg' must be tensor")
        shape = self.data_shape
        if arg.ndim != len(shape) + 1:
            raise ValueError(
                f"{type(self).__name__}.data() 'arg' must be {len(shape) + 1}-dimensional tensor"
            )
        shape = (arg.shape[0],) + shape
        if arg.shape != shape:
            raise ValueError(f"{type(self).__name__}.data() 'arg' must have shape {shape!r}")
        copy = shallow_copy(self)
        if callable(params):
            delattr(copy, "p")
        if isinstance(params, Parameter) and not isinstance(arg, Parameter):
            copy.params = Parameter(arg, params.requires_grad)
        else:
            copy.params = arg
        copy.clear_buffers()
        return copy

    def data_(
        self: Union[TSpatialTransform, ParametricTransform], arg: Tensor
    ) -> TSpatialTransform:
        r"""Replace transformation parameters.

        Args:
            arg: Tensor of transformation parameters with shape matching ``self.data_shape``,
                excluding the batch dimension whose size may be different from the current tensor.

        Returns:
            Reference to this in-place modified transformation module.

        Raises:
            ReadOnlyParameters: When ``self.params`` is a callable which provides the parameters.

        """
        params = self.params  # Note: May be None!
        if callable(params):
            raise ReadOnlyParameters(
                f"Cannot replace parameters, try {type(self).__name__}.data() instead."
            )
        if not isinstance(arg, Tensor):
            raise TypeError(f"{type(self).__name__}.data_() 'arg' must be tensor, not {type(arg)}")
        shape = self.data_shape
        if arg.ndim != len(shape) + 1:
            raise ValueError(
                f"{type(self).__name__}.data_() 'arg' must be {len(shape) + 1}-dimensional tensor"
                f", but arg.ndim={arg.ndim}"
            )
        shape = (arg.shape[0],) + shape
        if arg.shape != shape:
            raise ValueError(
                f"{type(self).__name__}.data_() 'arg' must have shape {shape!r}, not {arg.shape!r}"
            )
        if isinstance(params, Parameter) and not isinstance(arg, Parameter):
            self.params = Parameter(arg, params.requires_grad)
        else:
            self.params = arg
        self.clear_buffers()
        return self

    def _data(self: Union[TSpatialTransform, ParametricTransform]) -> Tensor:
        r"""Get most recent transformation parameters.

        When transformation parameters are obtained from a callable, this function invokes
        this callable with ``self.condition()`` as arguments if set, and returns the parameter
        obtained returned by this callable function or module. Otherwise, it simply returns a
        reference to the ``self.params`` tensor.

        Returns:
            Reference to ``self.params`` tensor or callable return value, respectively.

        """
        params = self.params
        if params is None:
            raise AssertionError(f"{type(self).__name__}._data() 'params' must be set first")
        if isinstance(params, type(self)):
            assert isinstance(params, ParametricTransform)
            return cast(ParametricTransform, params).data()
        if callable(params):
            args, kwargs = self.condition()
            pred = params(*args, **kwargs)
            if not isinstance(pred, Tensor):
                raise TypeError(f"{type(self).__name__}.params() value must be tensor")
            shape = self.data_shape
            if pred.ndim != len(shape) + 1:
                raise ValueError(
                    f"{type(self).__name__}.params() tensor must be {len(shape) + 1}-dimensional"
                )
            shape = (pred.shape[0],) + shape
            if pred.shape != shape:
                raise ValueError(f"{type(self).__name__}.params() tensor must have shape {shape!r}")
            return pred
        assert isinstance(params, Tensor)
        return params

    def link(
        self: Union[TSpatialTransform, ParametricTransform], other: TSpatialTransform
    ) -> TSpatialTransform:
        r"""Make shallow copy of this transformation which is linked to another instance."""
        return shallow_copy(self).link_(other)

    def link_(
        self: Union[TSpatialTransform, ParametricTransform],
        other: Union[TSpatialTransform, ParametricTransform],
    ) -> TSpatialTransform:
        r"""Link this transformation to another of the same type.

        This transformation is modified to use a reference to the given transformation. After linking,
        the transformation will not have parameters on its own, and its ``update()`` function will not
        recompute possibly previously shared parameters, e.g., parameters obtained by a callable neural
        network. Instead, it directly copies the parameters from the linked transformation.

        Args:
            other: Other transformation of the same type as ``self`` to which this transformation is linked.

        Returns:
            Reference to this transformation.

        """
        if other is self:
            raise ValueError(f"{type(self).__name__}.link() cannot link tranform to itself")
        if type(self) != type(other):
            raise TypeError(
                f"{type(self).__name__}.link() 'other' must be of the same type, got {type(other).__name__}"
            )
        self.params = other
        if not hasattr(self, "p"):
            if other.params is None:
                p = torch.empty(self.data_shape)
            else:
                p = other.data()
            self.register_buffer("p", p, persistent=False)
            if other.params is None:
                self.reset_parameters()
        return self

    def unlink(self: Union[TSpatialTransform, ParametricTransform]) -> TSpatialTransform:
        r"""Make a shallow copy of this transformation with parameters set to ``None``."""
        return shallow_copy(self).unlink_()

    def unlink_(self: Union[TSpatialTransform, ParametricTransform]) -> TSpatialTransform:
        r"""Resets transformation parameters to ``None``."""
        self.params = None
        if hasattr(self, "p"):
            delattr(self, "p")
        return self

    def update(self: Union[TSpatialTransform, ParametricTransform]) -> TSpatialTransform:
        r"""Update buffered data such as predicted parameters, velocities, and/or displacements."""
        if hasattr(self, "p"):
            p = self._data()
            self.register_buffer("p", p, persistent=False)
        super().update()
        return self


class InvertibleParametricTransform(ParametricTransform):
    r"""Mix-in for spatial transformations that support on-demand inversion."""

    def __init__(
        self,
        grid: Grid,
        groups: Optional[int] = None,
        params: Optional[Union[bool, Tensor, Callable[..., Tensor]]] = True,
        invert: bool = False,
    ) -> None:
        r"""Initialize transformation parameters.

        Args:
            grid: Grid domain on which transformation is defined.
            groups: Number of transformations. A given image batch can either be deformed by a
                single transformation, or a separate transformation for each image in the batch, e.g.,
                for group-wise or batched registration. The default is one transformation for all images
                in the batch, or the batch length of the ``params`` tensor if provided.
            params: Initial parameters. If a tensor is given, it is only registered as optimizable module
                parameters when of type ``torch.nn.Parameter``. When a callable is given instead, it will be
                called by ``self.update()`` with ``self.condition()`` arguments. When a boolean argument is
                given, a new zero-initialized tensor is created. If ``True``, this tensor is registered as
                optimizable module parameter.
            invert: Whether ``params`` correspond to the inverse transformation. When this flag is ``True``,
                the ``self.tensor()`` and related methods return the transformation corresponding to the
                inverse of the transformations with the given ``params``. For example in case of a rotation,
                the rotation matrix is first constructed from the rotation parameters (e.g., Euler angles),
                and then transposed if ``self.invert == True``. In general, inversion of linear transformations
                and non-rigid transformations parameterized by velocity fields can be done efficiently on-the-fly.

        """
        super().__init__(grid, groups=groups, params=params)
        self.invert = bool(invert)

    def inverse(
        self: Union[TSpatialTransform, InvertibleParametricTransform],
        link: bool = False,
        update_buffers: bool = False,
    ) -> TSpatialTransform:
        r"""Get inverse of this transformation.

        Args:
            link: Whether the inverse transformation keeps a reference to this transformation.
                If ``True``, the ``update()`` function of the inverse function will not recompute
                shared parameters (e.g., parameters obtained by a callable neural network), but
                directly access the parameters from this transformation. Note that when ``False``,
                the inverse transformation will still share parameters, modules, and buffers with
                this transformation, but these shared tensors may be replaced by a call of ``update()``
                (which is implicitly called as pre-forward hook when ``__call__()`` is invoked).
            update_buffers: Whether buffers of inverse transformation should be updated after creating
                the shallow copy. If ``False``, the ``update()`` function of the returned inverse
                transformation has to be called before it is used.

        Returns:
            Shallow copy of this transformation which computes and applies the inverse transformation.
            The inverse transformation will share the parameters with this transformation.

        """
        inv = shallow_copy(self)
        if link:
            inv.link_(self)
        inv.invert = not self.invert
        return inv

    def extra_repr(self: Union[TSpatialTransform, InvertibleParametricTransform]) -> str:
        r"""Print current transformation."""
        return super().extra_repr() + f", invert={self.invert}"



r"""Engine state output transformations."""

from typing import Callable, Optional, Sequence, Tuple, Union

from ignite.engine import Engine

import torch
from torch import Tensor

from deepali.core import ALIGN_CORNERS
from deepali.core.collections import TensorCollection, get_tensor
from deepali.core.image import grid_reshape
from deepali.core.tensor import as_one_hot_tensor


def get_output_transform(key: str) -> Callable[[TensorCollection], Tensor]:
    r"""Get tensor at specified nested engine output map key entry."""

    def output_transform(output: TensorCollection) -> Tensor:
        return get_tensor(output, key)

    return output_transform


def cm_binary_output_transform(
    y_pred: str = "y_pred",
    y: str = "y",
    reshape: str = "none",
    align_corners: bool = ALIGN_CORNERS,
) -> Callable[[TensorCollection], Tuple[Tensor, Tensor]]:
    r"""Convert engine output to a vector-valued one-hot prediction.

    Engine output transformation for use with ``ignite.metrics.ConfusionMatrix()``.

    Args:
        output: Ignite engine state output.

    Returns:
        y_pred: Tensor of one-hot encoded predictions with shape ``(N, 2, ..., X)``.
        y: Tensor of target labels with shape ``(N, ..., X)``.

    """

    def output_transform(output: TensorCollection) -> Tuple[Tensor, Tensor]:
        dtype = torch.int32
        y_pred_tensor = get_tensor(output, y_pred)
        y_tensor = get_tensor(output, y)
        assert y_tensor.ndim == y_pred_tensor.ndim
        assert y_tensor.shape[1] == 1
        if y_tensor.shape[0] == 1 and y_pred_tensor.shape[0] > 1:
            y_tensor = y_tensor.expand(y_pred_tensor.shape[0:1] + y_tensor.shape[1:])
        assert y_pred_tensor.shape[0] == y_tensor.shape[0]
        assert y_pred_tensor.shape[1] == 1
        grid_shape = y_tensor.shape[2:]
        if y_pred_tensor.shape[2:] != grid_shape and reshape == "y_pred":
            y_pred_tensor = grid_reshape(y_pred_tensor, grid_shape, align_corners=align_corners)
        y_pred_tensor = y_pred_tensor.round().type(torch.int64)
        y_pred_tensor = as_one_hot_tensor(y_pred_tensor, num_classes=2, dtype=dtype)
        if y_tensor.shape[2:] != grid_shape and reshape == "y":
            y_tensor = grid_reshape(y_tensor, grid_shape, align_corners=align_corners)
        y_tensor = y_tensor.round().type(dtype).squeeze(1)
        assert y_pred_tensor.shape[2:] == y_tensor.shape[1:]
        return y_pred_tensor, y_tensor

    return output_transform


def cm_multilabel_output_transform(
    y_pred: str = "y_pred", y: str = "y"
) -> Callable[[TensorCollection], Tuple[Tensor, Tensor]]:
    r"""Convert engine output to a vector-valued one-hot prediction.

    Engine output transformation for use with ``ignite.metrics.ConfusionMatrix()``.

    Args:
        output: Ignite engine state output.

    Returns:
        y_pred: Tensor of multi-class probabilites with shape ``(N, C, ..., X)``.
        y: Tensor of target labels with shape ``(N, ..., X)``.

    """

    def output_transform(output: TensorCollection) -> Tuple[Tensor, Tensor]:
        y_pred_tensor = get_tensor(output, y_pred)
        y_tensor = get_tensor(output, y)
        if y_tensor.shape[1] == 1:
            if y_tensor.is_floating_point():
                y_tensor = y_tensor.round().type(torch.int32)
            y_tensor = y_tensor.squeeze(1)
        else:
            y_tensor = y_tensor.argmax(dim=1)
        if y_tensor.shape[0] == 1 and y_pred_tensor.shape[0] > 1:
            y_tensor = y_tensor.expand(y_pred_tensor.shape[0:1] + y_tensor.shape[1:])
        return y_pred_tensor, y_tensor

    return output_transform


def cm_output_transform(
    y_pred: str = "y_pred",
    y: str = "y",
    multilabel: bool = False,
) -> Callable[[TensorCollection], Tuple[Tensor, Tensor]]:
    r"""Get engine output transformation for binary or multi-class segmentation output, respectively."""
    if multilabel:
        return cm_multilabel_output_transform(y_pred, y)
    return cm_binary_output_transform(y_pred, y)


def negative_loss_score_function(engine: Engine) -> Tensor:
    r"""Get negated loss value from ``engine.state.output``.

    This output transformation can be used as ``score_function`` argument of
    ``ignite.handlers.ModelCheckpoint``, for example.

    """
    output = engine.state.output
    if isinstance(output, dict):
        output = output["loss"]
    assert isinstance(output, Tensor)
    return -output


def y_pred_y_output_transform(
    y_pred: str = "y_pred",
    y: str = "y",
    channels: Optional[Union[int, Sequence[int]]] = None,
) -> Callable[[TensorCollection], Tuple[Tensor, Tensor]]:
    r"""Get engine output transformation which returns (y_pred, y) tuples.

    Args:
        y_pred: Output dictionary key for ``y_pred`` tensor.
        y: Output dictionary key for ``y`` tensor.
        channels: Indices of image channels to extract.

    """
    if isinstance(channels, int):
        channels = (channels,)
    if channels:

        def output_transform(output: TensorCollection) -> Tuple[Tensor, Tensor]:
            a = get_tensor(output, y_pred)
            b = get_tensor(output, y)
            return a[:, channels], b[:, channels]

    else:

        def output_transform(output: TensorCollection) -> Tuple[Tensor, Tensor]:
            a = get_tensor(output, y_pred)
            b = get_tensor(output, y)
            return a, b

    return output_transform


def y_pred_y_with_weight_output_transform(
    y_pred: str = "y_pred",
    y: str = "y",
    weight: str = "weight",
    kwarg: str = "weight",
    channels: Optional[Union[int, Sequence[int]]] = None,
) -> Callable[[TensorCollection], Tuple[Tensor, Tensor]]:
    r"""Get engine output transformation which returns (y_pred, y) tuples.

    Args:
        y_pred: Output dictionary key for ``y_pred`` tensor.
        y: Output dictionary key for ``y`` tensor.
        weight: Output dictionary key for ``weight`` tensor.
        kwarg: Name of weight keyword argument.
        channels: Indices of image channels to extract.

    """
    if isinstance(channels, int):
        channels = (channels,)
    if channels:

        def output_transform(output: TensorCollection) -> Tuple[Tensor, Tensor]:
            a = get_tensor(output, y_pred)
            b = get_tensor(output, y)
            w = get_tensor(output, weight)
            return a[:, channels], b[:, channels], {kwarg: w}

    else:

        def output_transform(output: TensorCollection) -> Tuple[Tensor, Tensor]:
            a = get_tensor(output, y_pred)
            b = get_tensor(output, y)
            w = get_tensor(output, weight)
            return a, b, {kwarg: w}

    return output_transform



r"""Auxiliary functions for working with vtkPolyData."""

from io import StringIO
from typing import List, Tuple

import numpy as np

from vtk import vtkCellArray, vtkPoints, vtkPolyData
from vtk import vtkOBJReader, vtkOBJWriter
from vtk import vtkPLYReader, vtkPLYWriter
from vtk import vtkXMLPolyDataReader, vtkXMLPolyDataWriter
from vtk import vtkPolyDataReader, vtkPolyDataWriter

from deepali.core.pathlib import PathUri
from deepali.core.storage import StorageObject

from .numpy import vtk_to_numpy_array, vtk_to_numpy_points


def read_polydata(path: PathUri) -> vtkPolyData:
    r"""Read vtkPolyData from specified file."""
    with StorageObject.from_path(path) as obj:
        if not obj.is_file():
            raise FileNotFoundError(f"File not found: {obj.uri}")
        obj = obj.pull(force=True)
        if not obj.path.is_file():
            raise FileNotFoundError(f"File not found: {obj.path}")
        suffix = obj.path.suffix.lower()
        if suffix == ".off":
            return read_polydata_off(obj.path)
        if suffix == ".obj":
            reader = vtkOBJReader()
        elif suffix == ".ply":
            reader = vtkPLYReader()
        elif suffix == ".vtp":
            reader = vtkXMLPolyDataReader()
        elif suffix == ".vtk":
            reader = vtkPolyDataReader()
        else:
            raise ValueError("Unsupported file name extension: {}".format(suffix))
        reader.SetFileName(str(obj.path))
        reader.Update()
        polydata = vtkPolyData()
        polydata.DeepCopy(reader.GetOutput())
    return polydata


def read_off(path: PathUri) -> Tuple[List[List[float]], List[List[int]]]:
    r"""Read values from .off file."""
    with StorageObject.from_path(path) as obj:
        data = obj.read_text()
    stream = StringIO(data)
    magic = stream.readline().strip()
    if magic not in ("OFF", "CNOFF"):
        raise ValueError(f"Invalid OFF file header: {path}")
    header = tuple([int(s) for s in stream.readline().strip().split(" ")])
    n_verts, n_faces = header[:2]
    verts = [[float(s) for s in stream.readline().strip().split(" ")] for _ in range(n_verts)]
    faces = [[int(s) for s in stream.readline().strip().split(" ")] for _ in range(n_faces)]
    assert len(verts) == n_verts, f"Expected {n_verts} vertices, found only {len(verts)}"
    assert len(faces) == n_faces, f"Expected {n_faces} vertices, found only {len(faces)}"
    return verts, faces


def read_polydata_off(path: PathUri) -> vtkPolyData:
    r"""Read vtkPolyData from .off file."""
    verts, faces = read_off(path)
    points = vtkPoints()
    polys = vtkCellArray()
    for vert in verts:
        points.InsertNextPoint(vert[:3])
    for poly in faces:
        polys.InsertNextCell(poly[0], poly[1 : 1 + poly[0]])
    output = vtkPolyData()
    output.SetPoints(points)
    output.SetPolys(polys)
    return output


def write_polydata(polydata: vtkPolyData, path: PathUri):
    r"""Write vtkPolyData to specified file in XML format."""
    with StorageObject.from_path(path) as obj:
        suffix = obj.path.suffix.lower()
        if suffix == ".off":
            write_polydata_off(polydata, obj.path)
            return
        if suffix == ".obj":
            writer = vtkOBJWriter()
        elif suffix == ".ply":
            writer = vtkPLYWriter()
            writer.SetFileTypeToBinary()
        elif suffix == ".vtp":
            writer = vtkXMLPolyDataWriter()
        elif suffix == ".vtk":
            writer = vtkPolyDataWriter()
        else:
            raise ValueError("Unsupported file name extension: {}".format(suffix))
        try:
            obj.path.unlink()  # in case of protected symlink to DVC cache
        except FileNotFoundError:
            obj.path.parent.mkdir(parents=True, exist_ok=True)
        writer.SetFileName(str(obj.path))
        writer.SetInputData(polydata)
        writer.Update()
        obj.push(force=True)


def write_polydata_off(polydata: vtkPolyData, path: PathUri):
    r"""Write vtkPolyData to specified file in OFF format."""
    verts = vtk_to_numpy_points(polydata)

    F = polydata.GetPolys().GetNumberOfCells()
    faces = vtk_to_numpy_array(polydata.GetPolys().GetData())
    assert faces.ndim == 1
    if len(faces) / F != 4:
        raise ValueError("write_polydata_off() only supports triangulated surface meshes")
    faces = faces.reshape(-1, 4)

    stream = StringIO()
    stream.write("OFF\n")
    stream.write(f"{len(verts)} {len(faces)} 0\n")
    np.savetxt(stream, verts, delimiter=" ", newline="\n", header="", footer="")
    np.savetxt(stream, faces, delimiter=" ", newline="\n", header="", footer="", fmt="%d")

    with StorageObject.from_path(path) as obj:
        obj.write_text(stream.getvalue())



r"""Interface for resources stored locally, remotely, or in cloud storage."""

from __future__ import annotations

from copy import deepcopy
from pathlib import Path
import re
import shutil
from typing import Generator, Optional, TypeVar, cast
from urllib.parse import urlsplit

from .pathlib import PathStr, PathUri, to_uri, unlink_or_mkdir


__all__ = (
    "LocalObject",
    "StorageObject",
    "copy_file",
)


TStorageObject = TypeVar("TStorageObject", bound="StorageObject")


class StorageObject(object):
    r"""Interface for storage objects.

    This base class can be used for storage objects that are only stored locally.
    The base implementations of the ``StorageObject`` interface functions reflect this
    use case, where ``StorageObject("/path/to/file")`` represents such local path object.
    The factory function ``StorageObject.from_uri`` is recommended for creating concrete
    instances of ``StorageObject`` or one of its subclasses. To create a resource instance
    for a local file path, use ``StorageObject.from_uri("file:///path/to/file")``.
    An S3 object resource is created by ``StorageObject.from_uri("s3://bucket/key")``.
    By using the ``StorageObject`` interface when writing tools that read and write
    from either local, remote, or cloud storage, the tool CLI can create these
    resource instances from input argument URIs or local file path strings
    without URI scheme, i.e., without "file://" prefix. The consumer or producer
    of a resource object can either directly read/write the object data using
    the ``read_(bytes|text)`` and/or ``write_(bytes|text)`` functions, or
    download/upload the storage object to/from a local file path using the
    ``pull`` and ``push`` operations. Note that these operations directly interact
    with the local storage if the resource instance is of base type ``StorageObject``,
    rather than a remote or cloud storage specific subclass. The ``pull`` and ``push``
    operations should be preferred over ``read`` and ``write`` if the resource data
    is accessed multiple times, in order to take advantage of the local temporary
    copy of the resource object. Otherwise, system IO operations can be saved by
    using the direct ``read`` and ``write`` operations instead.

    Additionally, the ``StorageObject.release()`` function should be called by tools
    when a resource is no longer required to indicate that the local copy of this
    resource can be removed. If the resource object itself represents a local
    ``StorageObject``, the release operation has no effect. To ensure that the ``release``
    function is called also in case of an exception, the ``StorageObject`` class
    implements the context manager interface functions ``__enter__`` and ``__exit__``.

    Example usage with storage object context:

    .. code-block:: python

        with StorageObject.from_uri("s3://bucket/key") as obj:
            # request download to local storage
            path = obj.pull().path
            # use local storage object referenced by path
        # local copy of storage object has been deleted

    The above is equivalent to using a try-finally block:

    .. code-block:: python

        obj = StorageObject.from_uri("s3://bucket/key")
        try:
            path = obj.pull().path
            # use local storage object referenced by path
        finally:
            # delete local copy of storage object
            obj.release()

    Usage of the ``with`` statement is recommended.

    Nesting of contexts for a resource object is possible and post-pones the
    invocation of the ``release`` operation until the outermost context has
    been left. This is accomplished by using a counter that is increment by
    ``__enter__``, and decremented again by ``__exit__``.

    It should be noted that ``StorageObject`` operations are generally not thread-safe,
    and actual consumers of resource objects should require the main thread to deal with
    obtaining, downloading (if ``pull`` is used), and releasing a resource. For different
    resources from remote storage (e.g., AWS S3), when using multiple processes (threads),
    the main process (thread) must initialize the default client connection (e.g., using
    ``S3Client.init_default()``) before spawning processes.

    """

    def __init__(self, path: PathStr) -> None:
        r"""Initialize storage object.

        Args:
            path (str, pathlib.Path): Local path of storage object.

        """
        self._path = Path(path).absolute()
        self._depth = 0

    def __enter__(self: TStorageObject) -> TStorageObject:
        r"""Enter context."""
        self._depth = max(1, self._depth + 1)
        return self

    def __exit__(self, *exc) -> None:
        r"""Release resource when leaving outermost context."""
        self._depth = max(0, self._depth - 1)
        if self._depth == 0:
            self.release()

    @staticmethod
    def from_path(*args: Optional[PathStr]) -> StorageObject:
        r"""Create storage object from path or URI.

        Args:
            args: Path or URI components. The last absolute path or URI is the base to which
                subsequent arguments are appended. Any ``None`` value is ignored. See also ``to_uri()``.

        Returns:
            obj (Resource): Instance of concrete type representing the referenced storage object.

        """
        return StorageObject.from_uri(to_uri(*args))

    @staticmethod
    def from_uri(uri: str) -> StorageObject:
        r"""Create storage object from URI.

        Args:
            uri: URI of storage object.

        Returns:
            Instance of concrete type representing the referenced storage object.

        """
        res = urlsplit(uri, scheme="file")
        if res.scheme == "file":
            match = re.match(r"/+([a-zA-Z]:.*)", res.path)
            path = match.group(1) if match else res.path
            return LocalObject(Path("/" + res.netloc + "/" + path if res.netloc else path))
        if res.scheme == "s3":
            # DO NOT import at module level to avoid cyclical import!
            try:
                from deepali.utils.aws.s3.object import S3Object
            except ImportError:
                raise ImportError("StorageObject.from_uri() requires AWS S3 deepali[utils]")
            return cast(StorageObject, S3Object.from_uri(uri))
        raise ValueError("Invalid or unsupported storage object URI: %s", uri)

    @property
    def uri(self) -> str:
        r"""
        Returns:
            uri (str): URI of storage object.

        """
        return self.path.as_uri()

    @property
    def path(self) -> Path:
        r"""Get absolute local path of storage object."""
        return self._path

    @property
    def name(self) -> str:
        r"""Name of storage object including file name extension, excluding directory path."""
        return self.path.name

    def with_path(self: TStorageObject, path) -> TStorageObject:
        r"""Create copy of storage object reference with modified ``path``.

        Args:
            path (str, pathlib.Path): New local path of storage object.

        Returns:
            New storage object reference with modified ``path`` property.

        """
        obj = deepcopy(self)
        obj._path = Path(path).absolute()
        return obj

    def with_properties(self: TStorageObject, **kwargs) -> TStorageObject:
        r"""Create copy of storage object reference with modified properties.

        Args:
            **kwargs: New property values. Only specified properties are changed.

        Returns:
            New storage object reference with modified properties.

        """
        obj = deepcopy(self)
        for name, value in kwargs.items():
            setattr(obj, name, value)
        return obj

    def exists(self) -> bool:
        r"""Whether object exists in storage."""
        return self.path.exists()

    def is_file(self) -> bool:
        r"""Whether storage object represents a file."""
        return self.path.is_file()

    def is_dir(self) -> bool:
        r"""Whether storage object represents a directory."""
        return self.path.is_dir()

    def iterdir(
        self: TStorageObject, prefix: Optional[str] = None
    ) -> Generator[TStorageObject, None, None]:
        r"""List storage objects within directory, excluding subfolder contents.

        Args:
            prefix: Name prefix.

        Returns:
            Generator of storage objects.

        """
        assert type(self) is StorageObject, "must be implemented by subclass"
        for path in self.path.iterdir():
            if not prefix or path.name.startswith(prefix):
                yield cast(TStorageObject, StorageObject(path))

    def pull(self: TStorageObject, force: bool = False) -> TStorageObject:
        r"""Download content of storage object to local path.

        Args:
            force (bool): Whether to force download even if local path already exists.

        Returns:
            This storage object.

        """
        return self

    def push(self: TStorageObject, force: bool = False) -> TStorageObject:
        r"""Upload content of local path to storage object.

        Args:
            force (bool): Whether to force upload even if storage object already exists.

        Returns:
            This storage object.

        """
        return self

    def read_bytes(self) -> bytes:
        r"""Read file content from local path if it exists, or referenced storage object otherwise.

        Returns:
            Binary file content of storage object.

        """
        return self.pull().path.read_bytes()

    def write_bytes(self: TStorageObject, data: bytes) -> TStorageObject:
        r"""Write bytes to storage object.

        Args:
            data: Binary data to write.

        Returns:
            This storage object.

        """
        try:
            self.path.unlink()
        except FileNotFoundError:
            self.path.parent.mkdir(parents=True, exist_ok=True)
        self.path.write_bytes(data)
        return self.push()

    def read_text(self, encoding: Optional[str] = None) -> str:
        r"""Read text file content from local path if it exists, or referenced storage object otherwise.

        Args:
            encoding: Text encoding.

        Returns:
            Decoded text file content of storage object.

        """
        return self.pull().path.read_text()

    def write_text(
        self: TStorageObject, text: str, encoding: Optional[str] = None
    ) -> TStorageObject:
        r"""Write text to storage object.

        Args:
            text: Text to write.
            encoding: Text encoding.

        Returns:
            This storage object.

        """
        try:
            self.path.unlink()
        except FileNotFoundError:
            self.path.parent.mkdir(parents=True, exist_ok=True)
        self.path.write_text(text, encoding=encoding)
        return self.push()

    def rmdir(self: TStorageObject) -> TStorageObject:
        r"""Remove directory both locally and from remote storage."""
        try:
            shutil.rmtree(self.path)
        except FileNotFoundError:
            pass
        return self

    def unlink(self: TStorageObject) -> TStorageObject:
        r"""Remove file both locally and from remote storage."""
        try:
            self.path.unlink()
        except FileNotFoundError:
            pass
        return self

    def delete(self: TStorageObject) -> TStorageObject:
        r"""Remove object both locally and from remote storage."""
        try:
            self.rmdir()
        except NotADirectoryError:
            self.unlink()
        return self

    def release(self: TStorageObject) -> TStorageObject:
        r"""Release local temporary copy of storage object.

        Only remove local copy of storage object. When the storage object
        is only stored locally, i.e., self is not a subclass of StorageObject,
        but of type ``StorageObject``, this operation does nothing.

        """
        if type(self) is not StorageObject:
            try:
                shutil.rmtree(self.path)
            except FileNotFoundError:
                pass
            except NotADirectoryError:
                try:
                    self.path.unlink()
                except FileNotFoundError:
                    pass
        return self

    def __str__(self) -> str:
        r"""Get human-readable string representation of storage object reference."""
        return self.uri

    def __repr__(self) -> str:
        r"""Get human-readable string representation of storage object reference."""
        return type(self).__name__ + "(path='{}')".format(self.path)


class LocalObject(StorageObject):
    r"""Local storage object."""

    def release(self: TStorageObject) -> TStorageObject:
        r"""Release local temporary copy of storage object."""
        return self


def copy_file(src: PathStr, dst: PathUri) -> StorageObject:
    r"""Copy local file to specified path or URI."""
    obj = StorageObject.from_uri(to_uri(dst))
    if isinstance(obj, LocalObject):
        path = unlink_or_mkdir(obj.path)
        shutil.copy2(src, path)
    else:
        obj.with_path(src).push(force=True)
    return obj



r"""Example implementation of free-form deformation (FFD) algorithm."""

import logging
from pathlib import Path
import sys
from timeit import default_timer as timer

import torch
import torch.cuda
from torch import Tensor

from deepali.core.argparse import ArgumentParser, Args, main_func
from deepali.core.config import read_config_dict
from deepali.core.environ import cuda_visible_devices
from deepali.core.grid import Grid
from deepali.core.logging import configure_logging
from deepali.core.pathlib import unlink_or_mkdir
from deepali.data import Image
from deepali.modules import TransformImage

from .pairwise import register_pairwise

log = logging.getLogger()


def parser(**kwargs) -> ArgumentParser:
    r"""Construct argument parser."""
    if "description" not in kwargs:
        kwargs["description"] = globals()["__doc__"]
    parser = ArgumentParser(**kwargs)
    parser.add_argument(
        "-c", "--config", help="Configuration file", default=Path(__file__).parent / "params.yaml"
    )
    parser.add_argument(
        "-t", "--target", "--target-img", dest="target_img", help="Fixed target image"
    )
    parser.add_argument(
        "-s", "--source", "--source-img", dest="source_img", help="Moving source image"
    )
    parser.add_argument("--target-seg", help="Fixed target segmentation label image")
    parser.add_argument("--source-seg", help="Moving source segmentation label image")
    parser.add_argument(
        "-o",
        "--output",
        "--output-transform",
        dest="output_transform",
        help="Output transformation parameters",
    )
    parser.add_argument(
        "-w",
        "--warped",
        "--warped-img",
        "--output-img",
        dest="warped_img",
        help="Deformed source image",
    )
    parser.add_argument(
        "--warped-seg",
        "--output-seg",
        dest="warped_seg",
        help="Deformed source segmentation label image",
    )
    parser.add_argument(
        "--device",
        help="Device on which to execute registration",
        choices=("cpu", "cuda"),
        default="cpu",
    )
    parser.add_argument("--debug-dir", help="Output directory for intermediate files")
    parser.add_argument(
        "--debug",
        "--debug-level",
        help="Debug level",
        type=int,
        default=0,
    )
    parser.add_argument("-v", "--verbose", help="Verbosity of output messages", type=int, default=0)
    parser.add_argument(
        "--log-level",
        help="Logging level",
        choices=["DEBUG", "INFO", "WARNING", "ERROR"],
        default="INFO",
    )
    return parser


def init(args: Args) -> int:
    r"""Initialize registration."""
    configure_logging(log, args)
    if args.device == "cuda":
        if not torch.cuda.is_available():
            log.error("Cannot use --device 'cuda' when torch.cuda.is_available() is False")
            return 1
        gpu_ids = cuda_visible_devices()
        if len(gpu_ids) != 1:
            log.error("CUDA_VISIBLE_DEVICES must be set to one GPU")
            return 1
    return 0


def func(args: Args) -> int:
    r"""Execute registration given parsed arguments."""
    config = read_config_dict(args.config)
    device = torch.device("cuda:0" if args.device == "cuda" else "cpu")
    start = timer()
    transform = register_pairwise(
        target={"img": args.target_img, "seg": args.target_seg},
        source={"img": args.source_img, "seg": args.source_seg},
        config=config,
        outdir=args.debug_dir,
        device=args.device,
        verbose=args.verbose,
        debug=args.debug,
    )
    log.info(f"Elapsed time: {timer() - start:.3f}s")
    if args.warped_img:
        target_grid = Grid.from_file(args.target_img)
        source_image = Image.read(args.source_img, device=device)
        warp_image = TransformImage(
            target=target_grid,
            source=source_image.grid(),
            sampling="linear",
            padding=source_image.min(),
        ).to(device)
        data: Tensor = warp_image(transform.tensor(), source_image)
        warped_image = Image(data, target_grid)
        warped_image.write(unlink_or_mkdir(args.warped_img))
    if args.warped_seg:
        target_grid = Grid.from_file(args.target_seg)
        source_image = Image.read(args.source_seg, device=device)
        warp_labels = TransformImage(
            target=target_grid,
            source=source_image.grid(),
            sampling="nearest",
            padding=0,
        ).to(device)
        data: Tensor = warp_labels(transform.tensor(), source_image)
        warped_image = Image(data, target_grid)
        warped_image.write(unlink_or_mkdir(args.warped_seg))
    if args.output_transform:
        path = unlink_or_mkdir(args.output_transform)
        if path.suffix == ".pt":
            transform.clear_buffers()
            torch.save(transform, path)
        else:
            transform.flow()[0].write(path)
    return 0


main = main_func(parser, func, init=init)


if __name__ == "__main__":
    sys.exit(main())



# %%
# Imports
from typing import Optional, Sequence

import matplotlib.pyplot as plt

import torch
from torch import Tensor
from torch.random import Generator

from deepali.core import Grid
import deepali.core.bspline as B
import deepali.core.functional as U


# %%
# Auxiliary functions
def random_svf(
    size: Sequence[int],
    stride: int = 1,
    generator: Optional[Generator] = None,
) -> Tensor:
    cp_grid_size = B.cubic_bspline_control_point_grid_size(size, stride=stride)
    data = torch.randn((1, 3) + cp_grid_size, generator=generator)
    data = U.fill_border(data, margin=3, value=0, inplace=True)
    return B.evaluate_cubic_bspline(data, size=size, stride=stride)


def visualize_flow(ax, flow: Tensor) -> None:
    grid = Grid(shape=flow.shape[2:], align_corners=True)
    x = grid.coords(channels_last=False, dtype=u.dtype, device=u.device)
    x = U.move_dim(x.unsqueeze(0).add_(flow), 1, -1)
    target_grid = U.grid_image(shape=flow.shape[2:], inverted=True, stride=(5, 5))
    warped_grid = U.warp_image(target_grid, x)
    ax.imshow(warped_grid[0, 0, flow.shape[2] // 2], cmap="gray")


# %%
# Random velocity fields
size = (128, 128, 128)
generator = torch.Generator().manual_seed(42)
u = random_svf(size, stride=8, generator=generator).mul_(0.1)
v = random_svf(size, stride=8, generator=generator).mul_(0.05)


# %%
# Evaluate displacement fields
flow_u = U.expv(u)
flow_v = U.expv(v)
flow = U.compose_flows(flow_u, flow_v)


# %%
# Approximate velocity field of composite displacement field
flow_w = U.expv(U.compose_svfs(u, v, bch_terms=3))


# %%
# Visualize composite displacement fields and error norm
fig, axes = plt.subplots(1, 3, figsize=(30, 10))

visualize_flow(axes[0], flow)
visualize_flow(axes[1], flow_w)

error = flow_w.sub(flow).norm(dim=1, keepdim=True)

ax = axes[2]
_ = ax.imshow(error[0, 0, error.shape[2] // 2], cmap="jet", vmin=0, vmax=0.1)


# %%



r"""Train Image-and-Spatial Transformer Network."""

from __future__ import annotations

import contextlib
from dataclasses import dataclass
from datetime import datetime
import logging
import math
from pathlib import Path
from typing import Any, Callable, Dict, Mapping, Optional, Sequence, Union

import typer
from typer import Option

from ignite import distributed as idist
from ignite.contrib.handlers.param_scheduler import LRScheduler
from ignite.engine import DeterministicEngine, Engine, Events
from ignite.metrics import Average, BatchWise, EpochWise, Metric, MetricUsage
from ignite.handlers import Checkpoint, DiskSaver, global_step_from_engine
from ignite.utils import manual_seed, setup_logger

import torch
import torch.backends.cudnn
from torch import Tensor
from torch.nn import Module, Sequential, functional as F
from torch.optim import Adam, Optimizer
from torch.optim.lr_scheduler import ExponentialLR
from torch.utils.data import DataLoader, Dataset, Subset
from torch.utils.data.distributed import DistributedSampler
from torch.utils.tensorboard import SummaryWriter

from deepali.core import functional as U
from deepali.core.config import DataclassConfig
from deepali.core.pathlib import unlink_or_mkdir
from deepali.data import ImageBatch, ImageDataset, ImageDatasetConfig, Partition
from deepali.data import collate_samples, prepare_batch
from deepali.data.transforms import CastImage, ResizeImage
from deepali.modules import ToImmutableOutput
from deepali.core.environ import cuda_visible_devices
from deepali.core.logging import LOG_FORMAT, LogLevel
from deepali.utils.ignite import handlers as H
from deepali.utils.ignite.output_transforms import get_output_transform
from deepali.utils.ignite.score_functions import negative_loss_score_function

from models import ImageAndSpatialTransformerConfig as ModelConfig
from models import ImageAndSpatialTransformerNetwork, create_istn


LossFunction = Callable[[Dict[str, Tensor]], Dict[str, Tensor]]


@dataclass
class DatasetsConfig(DataclassConfig):
    r"""Training dataset configuraiton."""

    train: ImageDatasetConfig
    valid: ImageDatasetConfig

    @classmethod
    def _from_dict(cls, arg: Mapping[str, Any], parent: Optional[Path] = None) -> Config:
        r"""Create configuration from dictionary."""
        if "train" not in arg:
            raise ValueError(
                f"{cls.__name__}.from_dict() 'arg' must contain 'train' dataset configuration"
            )
        if "valid" not in arg:
            raise ValueError(
                f"{cls.__name__}.from_dict() 'arg' must contain 'valid' dataset configuration"
            )
        train = ImageDatasetConfig.from_dict(arg["train"], parent=parent)
        valid = ImageDatasetConfig.from_dict(arg["valid"], parent=parent)
        return cls(train=train, valid=valid)


@dataclass
class TrainConfig(DataclassConfig):
    r"""Training hyperparameters."""

    # Initialization
    random_seed: int = 0
    deterministic: bool = True
    # Data loading
    batch_size: int = 10
    num_samples: int = -1
    num_workers_per_proc: int = 2
    num_workers_per_node: int = 10
    pin_memory: bool = True
    shuffle: bool = True
    # Loss function
    loss: str = "explicit"
    # Optimization parameters
    max_epochs: int = 100
    max_iterations: int = 0
    max_learning_rate: float = 0.01
    min_learning_rate: float = 0.0001
    learning_rate_decay_steps: int = 0
    learning_rate_decay_rate: float = 0.9
    # Logging settings
    log_every: int = 1
    checkpoint_every: int = 10
    checkpoint_lastn: int = 10
    checkpoint_score: str = "neg_loss"
    summary_every: int = 1
    summary_batches: bool = False
    summary_hists: bool = False
    summary_images: bool = True
    summary_graph: bool = False
    summary_grid: bool = True
    summary_grid_inverted: bool = True
    summary_grid_stride: Union[int, Sequence[int]] = (5, 5)
    # Evaluation on validation set
    eval_batch_size: int = -1
    eval_every_steps: int = 10
    eval_save_firstn: int = 5
    eval_num_samples: int = -1
    eval_num_workers: int = -1


@dataclass
class Config(DataclassConfig):
    r"""Dataset, model, and training configuration."""

    dataset: DatasetsConfig
    model: ModelConfig
    train: TrainConfig

    @classmethod
    def _from_dict(cls, arg: Mapping[str, Any], parent: Optional[Path] = None) -> Config:
        r"""Create configuration from dictionary."""
        if "dataset" not in arg:
            raise ValueError(
                f"{cls.__name__}.from_dict() 'arg' must contain 'dataset' configuration"
            )
        dataset = DatasetsConfig.from_dict(arg["dataset"], parent=parent)
        model = ModelConfig.from_dict(arg.get("model", {}), parent=parent)
        train = TrainConfig.from_dict(arg.get("train", {}), parent=parent)
        return cls(dataset=dataset, model=model, train=train)


def main(
    config_path: Path = Option(
        "params.yaml", "--config", "-c", exists=True, help="Configuration YAML file."
    ),
    log_dir: Path = Option("log/train", help="TensorBoard log directory."),
    log_level: LogLevel = Option(LogLevel.INFO, case_sensitive=False, help="Logging level name."),
    detect_anomaly: bool = Option(False, help="Enable anomaly detection."),
) -> None:
    r"""Invoke data parallel training."""

    now_dt = datetime.now().strftime("%Y-%m-%d_%H%M%S")
    logger = setup_logger("train", level=int(log_level), format=LOG_FORMAT, distributed_rank=0)

    # IDs of GPUs on which to execute training
    gpu_ids = cuda_visible_devices()
    if torch.cuda.device_count() != len(gpu_ids):
        logger.error(
            "CUDA device count (%d) must match environment variable CUDA_VISIBLE_DEVICES (%d)",
            torch.cuda.device_count(),
            len(gpu_ids),
        )
        return 1

    # Load configuration
    config = Config.read(config_path)

    # Save configuration to output log directory
    log_dir = str(log_dir).format(now=now_dt)
    log_dir: Path = Path(log_dir).absolute()
    log_dir.mkdir(parents=True, exist_ok=True)
    logger.debug(f"Write '{log_dir / 'train.yaml'}'")
    config.write(log_dir / "train.yaml")

    # Spawn processes for parallel model training
    spawn_kwargs = {}
    if len(gpu_ids) > 1:
        spawn_kwargs["backend"] = "nccl"
        spawn_kwargs["nproc_per_node"] = len(gpu_ids)
    with idist.Parallel(**spawn_kwargs) as parallel:
        parallel.run(
            train_local,
            config=config,
            log_dir=log_dir,
            log_level=log_level,
            now_dt=now_dt,
            detect_anomaly=detect_anomaly,
        )
    return 0


def train_local(
    local_rank: int,
    config: Config,
    log_dir: Path,
    log_level: Union[int, LogLevel],
    now_dt: str,
    trained_model_path: Optional[str] = None,
    detect_anomaly: bool = False,
) -> None:
    r"""Run model training in local distributed process."""

    rank = idist.get_rank()
    seed = config.train.random_seed + rank
    logger = setup_logger("train", level=int(log_level), format=LOG_FORMAT, distributed_rank=rank)

    logger.info(
        f"distributed backend={idist.backend()}, seed={seed}"
        f", rank={rank}, node_rank={idist.get_node_rank()}, local_rank={local_rank}"
    )
    logging.getLogger("ignite").setLevel(logging.WARNING)

    # Seed all random number generators
    manual_seed(seed)

    # Configure PyTorch
    torch.backends.cudnn.benchmark = False if config.train.deterministic else True
    torch.backends.cudnn.deterministic = config.train.deterministic
    torch.autograd.set_detect_anomaly(detect_anomaly)

    # Train ISTN
    if rank == 0 and log_dir:
        writer_cm = contextlib.closing(SummaryWriter(str(log_dir)))
    else:
        writer_cm = contextlib.nullcontext()
    with writer_cm as writer:
        istn = train(config, log_dir=log_dir, log_level=log_level, logger=logger, writer=writer)

    # Save best trained model
    if rank == 0 and trained_model_path:
        path = trained_model_path.format(logdir=log_dir.as_posix(), now=now_dt)
        path = unlink_or_mkdir(path)
        torch.save(istn.state_dict(), path)


def train(
    config: Config,
    load_model_path: Optional[Path] = None,
    checkpoint_path: Optional[Path] = None,
    log_dir: Optional[Path] = None,
    log_level: Union[int, LogLevel] = LogLevel.INFO,
    logger: Optional[logging.Logger] = None,
    writer: Optional[SummaryWriter] = None,
) -> ImageAndSpatialTransformerNetwork:
    r"""Train Image-and-Spatial Transformer Network."""

    rank = idist.get_rank()
    device = idist.device()

    if logger is None:
        logger = setup_logger(
            "train", level=int(log_level), format=LOG_FORMAT, distributed_rank=rank
        )

    # Initialize data loaders
    train_set = create_dataset(config, Partition.TRAIN)
    valid_set = create_dataset(config, Partition.VALID)

    if config.train.num_samples and config.train.num_samples > 0:
        train_set = Subset(train_set, list(range(config.train.num_samples)))
    if config.train.eval_num_samples and config.train.eval_num_samples > 0:
        valid_set = Subset(valid_set, list(range(config.train.eval_num_samples)))

    train_batches = create_dataloader(config, Partition.TRAIN, train_set)
    valid_batches = create_dataloader(config, Partition.VALID, valid_set)

    # Create and log model
    istn = create_istn(config.model)
    istn = istn.to(device=device)
    if rank == 0 and log_dir:
        model_txt = unlink_or_mkdir(log_dir / "model.txt")
        with model_txt.open("wt") as fp:
            print(istn, file=fp)
        if config.train.summary_graph and writer is not None:
            batch = next(iter(valid_batches))
            writer.add_graph(Sequential(istn, ToImmutableOutput()), [batch])
            del batch

    # Initialize parameters from pretrained model
    # before wrapping it in DistributedDataParallel
    is_pretrained = False
    if not checkpoint_path and load_model_path:
        if rank == 0:
            logger.info(f"Loading parameters from '{load_model_path}'")
        state_dict = torch.load(load_model_path, map_location=device)
        if "model" in state_dict:
            state_dict = state_dict["model"]
        if "state" in state_dict:
            state_dict = state_dict["state"]
        istn.load_state_dict(state_dict)
        is_pretrained = True

    # Wrap model in (Distributed)DataParallel
    model = idist.auto_model(istn)

    # Loss function
    loss = create_loss(config.train)

    # Setup optimizer
    optimizer = create_optimizer(config.train, model)
    optimizer = idist.auto_optim(optimizer)

    # Create training engine
    trainer: Engine = create_trainer(
        config.train, model, loss, optimizer, device=device, non_blocking=True
    )
    if isinstance(train_batches.sampler, DistributedSampler):
        trainer.add_event_handler(Events.EPOCH_STARTED, H.set_distributed_sampler_epoch)
    if config.train.log_every > 0:
        log_event = Events.ITERATION_COMPLETED(every=config.train.log_every)
        if config.train.log_every > 1:
            log_event |= Events.ITERATION_COMPLETED(once=1)
        trainer.add_event_handler(log_event, H.print_metrics, logger=logger)

    # Resume training from previous state and add handler to save checkpoints
    chkpt_objs = {"trainer": trainer, "model": model, "optimizer": optimizer}
    if checkpoint_path:
        if rank == 0:
            logger.info(f"Loading checkpoint '{checkpoint_path}'")
        chkpt = torch.load(checkpoint_path, map_location=device)
        Checkpoint.load_objects(to_load=chkpt_objs, checkpoint=chkpt)
    if config.train.checkpoint_every > 0:
        trainer.add_event_handler(
            Events.ITERATION_COMPLETED(every=config.train.checkpoint_every),
            checkpoint_writer(config.train, log_dir, chkpt_objs),
        )

    # Initial best state
    best_state_dict = model.state_dict()
    best_loss_value = float("inf")

    # Add handlers to save training summary to tensorboard event file
    if writer is not None and config.train.summary_every > 0:
        save_event = Events.ITERATION_COMPLETED(every=config.train.summary_every)
        if config.train.summary_every > 1:
            save_event |= Events.ITERATION_COMPLETED(once=1)
        trainer.add_event_handler(save_event, H.write_summary_metrics, writer, prefix="train/")
        if config.train.summary_batches:
            trainer.add_event_handler(
                save_event,
                H.write_summary_images,
                writer,
                prefix="train/",
                names={
                    "source_img",
                    "target_img",
                    "warped_img",
                    "source_seg",
                    "target_seg",
                    "warped_seg",
                    "source_soi",
                    "target_soi",
                    "warped_soi",
                    "warped_grid",
                },
                rescale_transform=normalize_summary_image,
            )
        if config.train.summary_hists:
            trainer.add_event_handler(save_event, H.write_summary_hists, writer, model)
        trainer.add_event_handler(
            save_event,
            H.write_summary_optim_params,
            writer,
            optimizer,
            prefix="optim/",
        )

    # Setup model evaluator
    eval_epoch_length = len(valid_batches)
    eval_every_steps = config.train.eval_every_steps
    if eval_every_steps < 0:
        eval_every_steps = config.train.summary_every
    if len(valid_batches) > 0 and eval_every_steps > 0:
        global_step_fn = global_step_from_engine(trainer)
        evaluator = create_evaluator(config.train, model, loss, device=device, non_blocking=True)
        if writer is not None:
            evaluator.add_event_handler(
                Events.COMPLETED,
                H.write_summary_metrics,
                writer,
                prefix="valid/",
                global_step_transform=global_step_fn,
            )
            if config.train.summary_images and config.train.eval_save_firstn != 0:

                def write_output_event_filter(_: Engine, event: int) -> bool:
                    if config.train.eval_save_firstn < 0:
                        return True
                    return event <= config.train.eval_save_firstn

                def write_input_event_filter(engine: Engine, event: int) -> bool:
                    if trainer.state.iteration > eval_every_steps:
                        return False
                    return write_output_event_filter(engine, event)

                prefix = "valid/{i}/"
                evaluator.add_event_handler(
                    Events.ITERATION_COMPLETED(write_input_event_filter),
                    H.write_summary_images,
                    writer,
                    prefix=prefix,
                    names={
                        "source_img",
                        "target_img",
                        "source_seg",
                        "target_seg",
                    },
                    rescale_transform=normalize_summary_image,
                    global_step_transform=global_step_fn,
                )
                evaluator.add_event_handler(
                    Events.ITERATION_COMPLETED(write_output_event_filter),
                    H.write_summary_images,
                    writer,
                    prefix=prefix,
                    names={
                        "source_soi",
                        "target_soi",
                        "warped_img",
                        "warped_seg",
                        "warped_soi",
                        "warped_grid",
                    },
                    rescale_transform=normalize_summary_image,
                    global_step_transform=global_step_fn,
                )
                del prefix

        eval_logger = setup_logger(
            "valid", level=int(log_level), format=LOG_FORMAT, distributed_rank=rank
        )
        evaluator.add_event_handler(
            Events.COMPLETED, H.print_metrics, logger=eval_logger, prefix=""
        )

        def evaluate(_: Engine):
            eval_logger.info(f"Evaluate metrics on validation set of size N={len(valid_set)}.")
            evaluator.run(valid_batches, max_epochs=1, epoch_length=eval_epoch_length)
            mean_loss = float(evaluator.state.metrics["loss"])
            nonlocal best_loss_value
            if mean_loss < best_loss_value:
                best_loss_value = mean_loss
                eval_logger.debug(
                    f"Copy state dict of new best model with average loss={mean_loss}."
                )
                nonlocal best_state_dict
                for name, value in model.state_dict().items():
                    best_state_dict[name].copy_(value)

        eval_event = Events.ITERATION_COMPLETED(every=eval_every_steps)
        if is_pretrained:
            eval_event |= Events.STARTED
        trainer.add_event_handler(eval_event, evaluate)

    # Run model training
    epoch_length = len(train_batches)
    max_epochs = config.train.max_epochs
    if config.train.max_iterations > 0:
        max_epochs = min(max_epochs, int(math.ceil(config.train.max_iterations / epoch_length)))
    trainer.run(train_batches, max_epochs=max_epochs, epoch_length=epoch_length)

    # Return trained model
    istn.load_state_dict(best_state_dict)
    return istn


def process_function(
    config: TrainConfig,
    istn: ImageAndSpatialTransformerNetwork,
    loss_fn: LossFunction,
    optimizer: Optional[Optimizer] = None,
    device: Optional[torch.device] = None,
    non_blocking: bool = False,
) -> Callable[[Engine, Dict[str, Union[str, ImageBatch, None]]], Dict[str, Tensor]]:
    r"""Process function of Ignite training or evaluation engine."""

    def detached(output: Dict[str, Tensor]) -> Dict[str, Tensor]:
        return {name: tensor.detach() for name, tensor in output.items()}

    def process_batch(_: Engine, batch: Dict[str, ImageBatch]) -> Dict[str, Tensor]:
        output: Dict[str, Tensor] = {}
        istn.train(optimizer is not None)
        with torch.set_grad_enabled(istn.training):
            # Copy input to device
            batch = prepare_batch(batch, device=device, non_blocking=non_blocking)
            output.update({k: v.tensor() for k, v in batch.items() if k != "meta"})
            # Evaluate ISTN output
            source_img = output["source_img"]
            target_img = output["target_img"]
            output.update(istn(source_img, target_img, apply=False))
            source_images = {k: v for k, v in output.items() if k.startswith("source_")}
            warped_images: Dict[str, Tensor] = istn.warp(source_images)
            output.update({k.replace("source_", "warped_"): v for k, v in warped_images.items()})
            # Render warped grid lines for visualization in TensorBoard
            if config.summary_grid:
                grid_image = U.grid_image(
                    shape=source_img.shape[2:],
                    stride=config.summary_grid_stride,
                    inverted=config.summary_grid_inverted,
                    device=device,
                )
                output["warped_grid"] = istn.warp(grid_image)
            # Evaluate loss terms
            losses = loss_fn(output)
            output.update(losses)
            if optimizer is not None:
                optimizer.zero_grad()
                losses["loss"].backward()
                optimizer.step()
        return detached(output)

    return process_batch


def attach_metrics(
    engine: Engine,
    usage: Union[MetricUsage, str] = EpochWise(),
) -> Engine:
    r"""Attach evaluation metrics to given Ignite engine."""

    loss = Average(get_output_transform("loss"))

    metrics: Dict[str, Metric] = {"loss": loss}
    # TODO: Additional metrics

    for name, metric in metrics.items():
        metric.attach(engine, name, usage=usage)
    return engine


def get_batch_size(config: TrainConfig, partition: Partition) -> int:
    batch_size = config.batch_size
    if partition is Partition.VALID and config.eval_batch_size > 0:
        batch_size = config.eval_batch_size
    return batch_size


def create_dataset(config: Config, partition: Union[Partition, str]) -> ImageDataset:
    partition = Partition(partition)
    if partition is Partition.TRAIN:
        dataset = ImageDataset.from_config(config.dataset.train)
    elif partition is Partition.VALID:
        dataset = ImageDataset.from_config(config.dataset.valid)
    else:
        raise ValueError(
            f"Dataset partition must be {Partition.TRAIN.value!r} or {Partition.VALID.value!r}"
        )
    keys = config.dataset.train.images.keys()
    transform = dataset.transform()
    transform.add_module("cast", CastImage.item(keys, dtype=torch.float32))
    transform.add_module("resize", ResizeImage.item(keys, size=config.model.input.size))
    return dataset


def create_dataloader(
    config: Config, partition: Union[Partition, str], dataset: Optional[Dataset] = None
) -> DataLoader:
    partition = Partition(partition)
    if dataset is None:
        dataset = create_dataset(config, partition)
    world_size = idist.get_world_size()
    shuffle = partition is Partition.TRAIN and config.train.shuffle
    if partition is Partition.TRAIN:
        num_workers = config.train.num_workers_per_proc
    else:
        num_workers = config.train.eval_num_workers
        if num_workers is None or num_workers < 0:
            num_workers = config.train.num_workers_per_proc
    batch_size = get_batch_size(config.train, partition)
    num_workers = max(0, min(num_workers * world_size, config.train.num_workers_per_node))
    persistent_workers = num_workers > 0 and partition is not Partition.TRAIN
    return idist.auto_dataloader(
        dataset,
        batch_size=batch_size * world_size,
        shuffle=shuffle,
        drop_last=False,
        pin_memory=config.train.pin_memory,
        num_workers=num_workers,
        persistent_workers=persistent_workers,
        collate_fn=collate_samples,
    )


def create_loss(config: TrainConfig) -> LossFunction:
    r"""Get loss function."""

    def loss_fn(output: Dict[str, Tensor]) -> Dict[str, Tensor]:
        r"""Evaluate loss terms given process function output."""
        # Loss terms used for progress monitoring
        loss_itn = F.mse_loss(output["source_soi"], output["source_seg"])
        loss_itn += F.mse_loss(output["target_soi"], output["target_seg"])
        loss_stn_u = F.mse_loss(output["warped_img"], output["target_img"])
        loss_stn_s = F.mse_loss(output["warped_seg"], output["target_seg"])
        loss_stn_i = F.mse_loss(output["warped_soi"], output["target_seg"])
        loss_stn_i += F.mse_loss(output["warped_seg"], output["target_soi"])
        loss_stn_r = F.mse_loss(output["warped_soi"], output["target_soi"])
        # Loss term used for training
        if config.loss in ("e", "explicit"):
            loss_train = loss_itn + loss_stn_s
        elif config.loss in ("i", "implicit"):
            loss_train = loss_stn_i + loss_stn_s
        elif config.loss in ("s", "supervised"):
            loss_train = loss_stn_s
        elif config.loss in ("u", "unsupervised"):
            loss_train = loss_stn_u
        else:
            raise ValueError(f"Invalid loss function: {config.loss}")
        # Return loss terms
        return {
            "loss": loss_train,
            "loss_itn": loss_itn,
            "loss_stn_u": loss_stn_u,
            "loss_stn_s": loss_stn_s,
            "loss_stn_i": loss_stn_i,
            "loss_stn_r": loss_stn_r,
        }

    return loss_fn


def create_optimizer(config: TrainConfig, model: Module) -> Optimizer:
    return Adam(model.parameters(), lr=config.max_learning_rate)


def create_trainer(
    config: TrainConfig,
    model: ImageAndSpatialTransformerNetwork,
    loss_fn: LossFunction,
    optimizer: Optimizer,
    device: Optional[Union[torch.device, str]] = None,
    non_blocking: bool = False,
) -> Engine:
    r"""Create Ignite engine for model training."""
    process_fn = process_function(
        config, model, loss_fn, optimizer=optimizer, device=device, non_blocking=non_blocking
    )
    engine_type = DeterministicEngine if config.deterministic else Engine
    trainer = engine_type(process_fn)
    if config.max_iterations and config.max_iterations > 0:
        trainer.add_event_handler(
            Events.ITERATION_COMPLETED,
            H.terminate_on_max_iteration,
            config.max_iterations,
        )
    if config.learning_rate_decay_steps > 0:
        if 0 < config.learning_rate_decay_rate < 1:
            lr_gamma = config.learning_rate_decay_rate ** (1 / config.learning_rate_decay_steps)
            lr_scheduler = LRScheduler(ExponentialLR(optimizer, lr_gamma))
            trainer.add_event_handler(Events.ITERATION_COMPLETED, lr_scheduler)
        else:
            raise ValueError(
                "create_trainer() 'config' learning rate decay rate must be in open interval (0, 1)."
                " To disable learning rate decay, set learning_rate_decay_steps=0."
            )
        if config.min_learning_rate > 0:
            trainer.add_event_handler(
                Events.ITERATION_STARTED,
                H.clamp_learning_rate,
                optimizer,
                min_learning_rate=config.min_learning_rate,
                max_learning_rate=config.max_learning_rate,
            )
    return attach_metrics(trainer, usage=BatchWise())


def create_evaluator(
    config: TrainConfig,
    model: ImageAndSpatialTransformerNetwork,
    loss_fn: LossFunction,
    device: Optional[torch.device] = None,
    non_blocking: bool = False,
) -> Engine:
    r"""Create Ignite engine for model evaluation."""
    process_fn = process_function(config, model, loss_fn, device=device, non_blocking=non_blocking)
    engine_type = DeterministicEngine if config.deterministic else Engine
    evaluator = engine_type(process_fn)
    return attach_metrics(evaluator, usage=EpochWise())


def checkpoint_writer(
    config: TrainConfig,
    log_dir: Optional[str],
    to_save: Mapping[str, Any],
    filename_pattern: Optional[str] = None,
    global_step_transform: Optional[Callable] = None,
) -> Checkpoint:
    r"""Create checkpoint handler."""
    n_saved = config.checkpoint_lastn
    if n_saved == 0 or not log_dir:
        return
    if n_saved is not None and n_saved < 0:
        n_saved = None
    score_function = None
    score_name = config.checkpoint_score
    if score_name == "neg_loss":
        score_function = negative_loss_score_function
    elif score_name:
        raise ValueError(f"checkpoint_writer() invalid 'config.checkpoint_score': {score_name}")
    if not filename_pattern:
        filename_pattern = "checkpoint"
        if score_name:
            filename_pattern += "_{score_name}={score}"
        else:
            filename_pattern += "_{global_step:06d}"
        filename_pattern += ".{ext}"
    disk_saver = DiskSaver(
        log_dir,
        atomic=True,
        create_dir=True,
        require_empty=False,
    )
    checkpoint_writer = Checkpoint(
        to_save=to_save,
        save_handler=disk_saver,
        score_function=score_function,
        score_name=score_name,
        n_saved=n_saved,
        global_step_transform=global_step_transform,
        filename_pattern=filename_pattern,
    )
    return checkpoint_writer


def normalize_summary_image(tag: str, data: Tensor) -> Tensor:
    r"""Linearly rescale image values to [0, 1] for logging to TensorBoard."""
    if tag.endswith("_seg"):
        return data
    return U.normalize_image(data, mode="unit")


if __name__ == "__main__":
    typer.run(main)
